package com.github.polpetta.mezzotre.i18n;

import java.util.Locale;
import java.util.ResourceBundle;
import javax.inject.Inject;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.runtime.resource.loader.JarResourceLoader;
import org.apache.velocity.tools.Scope;
import org.apache.velocity.tools.ToolContext;
import org.apache.velocity.tools.ToolManager;
import org.apache.velocity.tools.config.FactoryConfiguration;
import org.apache.velocity.tools.config.ToolConfiguration;
import org.apache.velocity.tools.config.ToolboxConfiguration;

/**
 * This class provides utility to create {@link ToolManager} for {@link VelocityEngine} or to
 * retrieve simple localized strings from the system.
 *
 * @author Davide Polonio
 * @since 1.0
 */
public class LocalizedMessageFactory {

  private static final String MESSAGE_PATH = "i18n/message";

  private final VelocityEngine velocityEngine;

  @Inject
  public LocalizedMessageFactory(VelocityEngine velocityEngine) {
    this.velocityEngine = velocityEngine;
  }

  /**
   * Provide a {@link ToolManager} completely setup. Localization will be taken from jar resources,
   * and {@link LocalizedTool} can be used in Velocity templates to automatically access them.
   *
   * @param locale the language needed
   * @return a {@link ToolManager} ready with the desired locale, if it exists. Fallback to english
   *     otherwise
   */
  public ToolManager createVelocityToolManager(Locale locale) {

    final ToolManager toolManager = new ToolManager();
    toolManager.setVelocityEngine(velocityEngine);
    final FactoryConfiguration factoryConfiguration = new FactoryConfiguration();
    final ToolboxConfiguration toolboxConfiguration = new ToolboxConfiguration();
    toolboxConfiguration.setScope(Scope.REQUEST);
    final ToolConfiguration toolConfiguration = new ToolConfiguration();
    toolConfiguration.setClassname(LocalizedTool.class.getName());
    toolConfiguration.setProperty("file.resource.loader.class", JarResourceLoader.class.getName());
    toolConfiguration.setProperty(ToolContext.LOCALE_KEY, locale);
    toolConfiguration.setProperty(LocalizedTool.BUNDLES_KEY, MESSAGE_PATH);
    toolboxConfiguration.addTool(toolConfiguration);
    factoryConfiguration.addToolbox(toolboxConfiguration);
    toolManager.configure(factoryConfiguration);
    return toolManager;
  }

  /**
   * Creates a {@link ResourceBundle} object that can be used to access single localized strings
   *
   * @param locale the language desired to retrieve the translated strings
   * @return a {@link ResourceBundle} with the provided locale
   */
  public ResourceBundle createResourceBundle(Locale locale) {
    return ResourceBundle.getBundle(MESSAGE_PATH, locale);
  }
}
