package com.github.polpetta.mezzotre.i18n;

import java.nio.charset.StandardCharsets;
import java.util.Locale;
import java.util.function.Consumer;
import javax.inject.Inject;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.tools.ToolManager;
import org.apache.velocity.util.StringBuilderWriter;

// FIXME tests

/**
 * This class aims to generate localized messages or strings either merging Velocity templates or
 * taking directly the keys from {@link java.util.ResourceBundle}
 *
 * @author Davide Polonio
 * @since 1.0
 */
public class TemplateContentGenerator {

  private final LocalizedMessageFactory localizedMessageFactory;

  @Inject
  public TemplateContentGenerator(LocalizedMessageFactory localizedMessageFactory) {
    this.localizedMessageFactory = localizedMessageFactory;
  }

  /**
   * Merge a Velocity template located in the local Jar resource path with the given {@link Locale},
   * provided as a {@link String}. A {@link Consumer} is provided in order to let the user customize
   * the {@link VelocityContext} with additional variables. Localization strings will be directly
   * taken via the use of {@link LocalizedTool} class.
   *
   * @param velocityContextConsumer a lambda function that allows for the customization of {@link
   *     VelocityContext} instance
   * @param localeAsString a {@link Locale} formatted as a {@link String}. Note that {@link
   *     Locale#forLanguageTag(String)} will be used for the conversion
   * @param templateName the path to the local Jar resource
   * @return a {@link String} containing the localized message
   */
  public String mergeTemplate(
      Consumer<VelocityContext> velocityContextConsumer,
      String localeAsString,
      String templateName) {
    final Locale locale = Locale.forLanguageTag(localeAsString);
    final ToolManager toolManager = localizedMessageFactory.createVelocityToolManager(locale);
    final VelocityContext velocityContext = new VelocityContext(toolManager.createContext());

    velocityContextConsumer.accept(velocityContext);

    final StringBuilder content = new StringBuilder();
    final StringBuilderWriter stringBuilderWriter = new StringBuilderWriter(content);

    toolManager
        .getVelocityEngine()
        .mergeTemplate(
            templateName, StandardCharsets.UTF_8.name(), velocityContext, stringBuilderWriter);

    stringBuilderWriter.close();
    return content.toString();
  }

  /**
   * Get a localized {@link String}
   *
   * @param locale a provided {@link Locale}
   * @param key the key that will be retrieved from the localized messages
   * @return a localized {@link String}
   * @see #getString(String, String)
   */
  public String getString(Locale locale, String key) {
    return localizedMessageFactory.createResourceBundle(locale).getString(key);
  }

  /**
   * Get a localized {@link String}
   *
   * @param localeAsString a provided {@link Locale} as a {@link String}. Will be converted using
   *     {@link Locale#forLanguageTag(String)}
   * @param key the key that will be retrieved from the localized messages
   * @return a localized {@link String}
   * @see #getString(Locale, String)
   */
  public String getString(String localeAsString, String key) {
    return getString(Locale.forLanguageTag(localeAsString), key);
  }
}
