package com.github.polpetta.mezzotre.orm;

import com.google.common.util.concurrent.AbstractExecutionThreadService;
import io.ebean.typequery.PString;
import io.ebean.typequery.TQRootBean;
import io.vavr.Tuple;
import io.vavr.Tuple3;
import io.vavr.control.Try;
import java.util.LinkedList;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;
import javax.inject.Inject;
import javax.inject.Named;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;

/**
 * This class works as a "batch" persistence entries remover - basically, it removes all the entries
 * that are in the database in an async way. This allows better performance on other pieces of
 * codebase, where we don't really care about checking that the removal has been successful, and
 * instead we just want to issue a deletion and instantly continue with our codebase flow.
 *
 * <p>This service provides the ability to listen to the operation in two ways: via old-fashioned
 * listeners and via {@link CompletableFuture}. If the caller wants to be sure that the removal has
 * effectively happened, it can sync with the given {@link CompletableFuture}.
 *
 * @author Davide Polonio
 * @since 1.0
 */
public class BatchBeanCleanerService extends AbstractExecutionThreadService {

  private final LinkedBlockingDeque<
          Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>
      entriesToRemove;
  private final Logger log;
  private final Pair<Integer, TimeUnit> serviceRunningCheckTime;

  // Accessing adding listeners while we iterate on the list inside the service thread _can_ be
  // considered a race condition - but given the frequency of adding and removing listeners, the
  // computational cost of putting Locks in place and check for them everytime is way greater
  private final LinkedList<
          Consumer<Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>>
      deletionListener;

  @Inject
  public BatchBeanCleanerService(
      Logger logger,
      @Named("serviceRunningCheckTime") Pair<Integer, TimeUnit> serviceRunningCheckTime) {
    this.log = logger;
    this.serviceRunningCheckTime = serviceRunningCheckTime;
    this.entriesToRemove = new LinkedBlockingDeque<>();
    this.deletionListener = new LinkedList<>();
  }

  @Override
  protected void run() throws Exception {
    log.trace("BatchBeanCleanerService run() method invoked");
    while (isRunning()) {
      // This statement is blocking for 1 sec if the queue is empty, and then it goes on - this way
      // we check if the service is still supposed to be up or what
      Optional.ofNullable(
              entriesToRemove.poll(
                  serviceRunningCheckTime.getLeft(), serviceRunningCheckTime.getRight()))
          .ifPresent(
              entryToRemove -> {
                Try.of(
                        () -> {
                          final int deleted = entryToRemove._2().eq(entryToRemove._1()).delete();
                          if (deleted > 0) {
                            log.trace(
                                "Bean with id " + entryToRemove._1() + " removed successfully ");
                          } else {
                            log.warn(
                                "Bean(s) with id "
                                    + entryToRemove._1()
                                    + " for query "
                                    + entryToRemove._2()
                                    + " was not removed from the database because it was not"
                                    + " found");
                          }
                          entryToRemove._3().complete(deleted);
                          deletionListener.parallelStream().forEach(l -> l.accept(entryToRemove));
                          return null;
                        })
                    .onFailure(ex -> entryToRemove._3().completeExceptionally(ex));
              });
    }
  }

  @Override
  protected void shutDown() throws Exception {
    super.shutDown();
    entriesToRemove.clear();
  }

  /**
   * Add an entry to be removed
   *
   * @param id the id of the entry to remove
   * @param column the column that will be used to perform the delete statement
   * @return a {@link CompletableFuture} containing an {@link Integer} indicating the number of rows
   *     affected by the operation
   */
  public CompletableFuture<Integer> removeAsync(
      String id, PString<? extends TQRootBean<?, ?>> column) {
    final CompletableFuture<Integer> jobExecution = new CompletableFuture<>();
    entriesToRemove.offer(Tuple.of(id, column, jobExecution));
    return jobExecution;
  }

  /**
   * Add a listener that will be invoked once the entry has been removed. This listener is invoked
   * after the {@link CompletableFuture} completion. Note that all the listeners are called in a
   * parallel fashion, so call order can change any time.
   *
   * @param listener the listener to add.
   */
  public void addListener(
      Consumer<Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>
          listener) {
    deletionListener.add(listener);
  }

  /**
   * Remove a listener
   *
   * @param listener the listener to be removed
   */
  public void removeListener(
      Consumer<Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>
          listener) {
    deletionListener.remove(listener);
  }
}
