package com.github.polpetta.mezzotre.orm.telegram;

import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.github.polpetta.mezzotre.util.Clock;
import com.github.polpetta.types.json.ChatContext;
import com.pengrad.telegrambot.model.Message;
import com.pengrad.telegrambot.model.Update;
import java.util.Collections;
import java.util.Map;
import java.util.Optional;
import javax.inject.Inject;

/**
 * ChatUtil provides utilities for interacting in an easier way when manipulating chat related data.
 * In particular, it provides an easy and DRY way to modify a set of frequently-accessed fields, or
 * simply to extract data performing all the necessary steps
 *
 * @author Davide Polonio
 * @since 1.0
 */
public class ChatUtil {

  private final Clock clock;

  @Inject
  public ChatUtil(Clock clock) {
    this.clock = clock;
  }

  /**
   * Update the {@link ChatContext} with the values passed to the method. The updated values are
   * then saved in the database
   *
   * @param chat the chat that will be updated with the new {@link ChatContext} values
   * @param stepName the step name to set
   * @param stageNumber the stage number to set
   * @param additionalFields if there are, additional custom fields that will be added to {@link
   *     ChatContext}. Note that these values will have to be manually retrieved since no method is
   *     available for custom entries. Use {@link Collections#emptyMap()} if you don't wish to add
   *     any additional field
   */
  public void updateChatContext(
      TgChat chat, String stepName, int stageNumber, Map<String, Object> additionalFields) {
    final ChatContext chatContext = chat.getChatContext();
    chatContext.setStage(stepName);
    chatContext.setStep(stageNumber);
    chatContext.setPreviousMessageUnixTimestampInSeconds(clock.now());
    additionalFields.forEach(chatContext::setAdditionalProperty);

    chat.setChatContext(chatContext);
    chat.save();
  }

  /**
   * Retrieves a possible {@link TgChat} from the given {@link CallbackQueryContext} or the {@link
   * Update}. Note that first the {@link CallbackQueryContext} is checked, then {@link Update} is
   * used as fallback. Once the Telegram chat id is retrieved from one of these two, the persistence
   * layer is queried and if there is any {@link TgChat} which the corresponding ID it is returned
   * to the caller.
   *
   * @param callbackQueryContext the {@link CallbackQueryContext} coming from a Telegram callback
   *     query
   * @param update the whole {@link Update} object that is sent from Telegram servers
   * @return an {@link Optional} that may contain a {@link TgChat} if the ID is found in the
   *     persistence layer, otherwise {@link Optional#empty()} is given back either if the id is not
   *     present or if the chat is not present in the persistence layer.
   */
  public Optional<TgChat> extractChat(CallbackQueryContext callbackQueryContext, Update update) {
    return Optional.of(callbackQueryContext.getFields().getTelegramChatId())
        .map(Double::longValue)
        // If we're desperate, search in the message for the chat id
        .filter(chatId -> chatId != 0L && chatId != Long.MIN_VALUE)
        .or(
            () ->
                Optional.ofNullable(update.callbackQuery().message())
                    .map(Message::messageId)
                    .map(Long::valueOf))
        .filter(chatId -> chatId != 0L && chatId != Long.MIN_VALUE)
        .flatMap(chatId -> new QTgChat().id.eq(chatId).findOneOrEmpty());
  }
}
