package com.github.polpetta.mezzotre.telegram.callbackquery;

import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.request.BaseRequest;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;

/**
 * This class dispatches incoming {@link Update} containing a {@link
 * com.pengrad.telegrambot.model.CallbackQuery} entry to the right {@link Processor} that will take
 * care of processing them.
 *
 * @author Davide Polonio
 * @since 1.0
 */
@Singleton
public class Dispatcher {

  private final Map<String, Processor> tgEventProcessors;
  private final Executor threadPool;
  private final NotFoundFactory notFoundFactory;

  @Inject
  public Dispatcher(
      @Named("eventProcessors") Map<String, Processor> tgEventProcessors,
      @Named("eventThreadPool") Executor threadPool,
      NotFoundFactory notFoundFactory) {
    this.tgEventProcessors = tgEventProcessors;
    this.threadPool = threadPool;
    this.notFoundFactory = notFoundFactory;
  }

  /**
   * This method searches for the right {@link Processor} installed in the system. A {@link
   * CallbackQueryContext} has to be previously saved in the database in order for this method to
   * find the right event to process. Once the corresponding {@link Processor} is found, the
   * computation is delegated to it
   *
   * @param callbackQueryContext the context coming from the database storage - cannot be null
   * @param update the incoming {@link Update} coming from Telegram
   * @return a {@link CompletableFuture} containing an {@link Optional} that may or not have a
   *     {@link BaseRequest} response to send back to Telegram.
   */
  public CompletableFuture<Optional<BaseRequest<?, ?>>> dispatch(
      CallbackQueryContext callbackQueryContext, Update update) {
    return CompletableFuture.completedFuture(update)
        .thenComposeAsync(
            ignored ->
                Optional.of(callbackQueryContext.getFields().getEvent())
                    .map(
                        eventName ->
                            tgEventProcessors
                                .getOrDefault(eventName, notFoundFactory.create(eventName))
                                .process(callbackQueryContext, update))
                    .orElse(CompletableFuture.failedFuture(new EventProcessorNotFoundException())),
            threadPool);
  }
}
