package com.github.polpetta.mezzotre.telegram.callbackquery;

import com.github.polpetta.mezzotre.orm.CallbackQueryContextCleaner;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.telegram.model.Help;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.model.request.ParseMode;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.EditMessageText;
import com.pengrad.telegrambot.request.SendMessage;
import java.util.Collections;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;

/**
 * ShowHelp callback query event edits a previous message (if available), otherwise it sends a new
 * message with a list of available commands (the one implemented by {@link
 * com.github.polpetta.mezzotre.telegram.command.Processor}), and a list of buttons that are
 * callback queries (the ones implemented by {@link Processor}). The class is a Singleton since it
 * is stateless and having multiple instances would only be a waste of memory.
 *
 * @author Davide Polonio
 * @since 1.0
 * @see com.github.polpetta.mezzotre.telegram.command.Help same functionality but as {@link
 *     com.github.polpetta.mezzotre.telegram.command.Processor} command
 */
@Singleton
class ShowHelp implements Processor {

  public static String EVENT_NAME = "showHelp";
  private final Executor threadPool;
  private final Help modelHelp;
  private final Map<String, com.github.polpetta.mezzotre.telegram.command.Processor>
      tgCommandProcessors;
  private final Map<String, Processor> eventProcessor;
  private final ChatUtil chatUtil;
  private final CallbackQueryContextCleaner callbackQueryContextCleaner;

  @Inject
  public ShowHelp(
      @Named("eventThreadPool") Executor threadPool,
      com.github.polpetta.mezzotre.telegram.model.Help modelHelp,
      @Named("commandProcessor")
          Map<String, com.github.polpetta.mezzotre.telegram.command.Processor> tgCommandProcessors,
      @Named("eventProcessors")
          Map<String, com.github.polpetta.mezzotre.telegram.callbackquery.Processor> eventProcessor,
      ChatUtil chatUtil,
      CallbackQueryContextCleaner callbackQueryContextCleaner) {
    this.threadPool = threadPool;
    this.modelHelp = modelHelp;
    this.tgCommandProcessors = tgCommandProcessors;
    this.eventProcessor = eventProcessor;
    this.chatUtil = chatUtil;
    this.callbackQueryContextCleaner = callbackQueryContextCleaner;
  }

  @Override
  public String getEventName() {
    return EVENT_NAME;
  }

  @Override
  public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
      CallbackQueryContext callbackQueryContext, Update update) {
    return CompletableFuture.supplyAsync(
        () ->
            chatUtil
                .extractChat(callbackQueryContext, update)
                .map(
                    chat -> {
                      final String message = modelHelp.getMessage(chat, tgCommandProcessors);
                      chatUtil.updateChatContext(chat, EVENT_NAME, 0, Collections.emptyMap());
                      chat.setHasHelpBeenShown(true);
                      chat.save();
                      final Optional<Integer> messageId = Util.extractMessageId(update);
                      final InlineKeyboardButton[] buttons =
                          modelHelp.generateInlineKeyBoardButton(chat, eventProcessor);
                      BaseRequest<?, ?> request;
                      if (messageId.isPresent()) {
                        final EditMessageText editMessageText =
                            new EditMessageText(chat.getId(), messageId.get(), message)
                                .parseMode(ParseMode.Markdown);
                        if (buttons.length > 0) {
                          editMessageText.replyMarkup(new InlineKeyboardMarkup(buttons));
                        }
                        request = editMessageText;
                      } else {
                        final SendMessage sendMessage =
                            new SendMessage(chat.getId(), message).parseMode(ParseMode.Markdown);
                        if (buttons.length > 0) {
                          sendMessage.replyMarkup(new InlineKeyboardMarkup(buttons));
                        }
                        request = sendMessage;
                      }

                      // We don't check if the element is removed or what - we just schedule its
                      // removal, then it is someone else problem
                      callbackQueryContextCleaner.removeGroupAsync(
                          callbackQueryContext.getEntryGroup());

                      return request;
                    }),
        threadPool);
  }
}
