package com.github.polpetta.mezzotre.telegram.command;

import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.model.request.ParseMode;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import java.util.Collections;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;

/**
 * The Help command class allows the user to be informed in any moment of commands that are
 * available in the system. If possible, it also provides a {@code reply_markup} keyboard ({@link
 * InlineKeyboardMarkup} for easier bot interactions
 *
 * @author Davide Polonio
 * @since 1.0
 * @see com.github.polpetta.mezzotre.telegram.callbackquery.ShowHelp for event-based help
 */
@Singleton
class Help implements Processor {

  private static final String TRIGGERING_STAGING_NAME = "/help";

  private final Executor threadPool;
  private final Map<String, Processor> tgCommandProcessors;
  private final Map<String, com.github.polpetta.mezzotre.telegram.callbackquery.Processor>
      eventProcessor;
  private final com.github.polpetta.mezzotre.telegram.model.Help modelHelp;
  private final ChatUtil chatUtil;

  @Inject
  public Help(
      @Named("eventThreadPool") Executor threadPool,
      @Named("commandProcessor") Map<String, Processor> tgCommandProcessors,
      @Named("eventProcessors")
          Map<String, com.github.polpetta.mezzotre.telegram.callbackquery.Processor> eventProcessor,
      com.github.polpetta.mezzotre.telegram.model.Help modelHelp,
      ChatUtil chatUtil) {
    this.threadPool = threadPool;
    this.tgCommandProcessors = tgCommandProcessors;
    this.eventProcessor = eventProcessor;
    this.modelHelp = modelHelp;
    this.chatUtil = chatUtil;
  }

  @Override
  public Set<String> getTriggerKeywords() {
    return Set.of(TRIGGERING_STAGING_NAME);
  }

  @Override
  public CompletableFuture<Optional<BaseRequest<?, ?>>> process(TgChat chat, Update update) {
    return CompletableFuture.supplyAsync(
        () -> {
          final String message = modelHelp.getMessage(chat, tgCommandProcessors);
          chatUtil.updateChatContext(chat, TRIGGERING_STAGING_NAME, 0, Collections.emptyMap());
          chat.setHasHelpBeenShown(true);
          chat.save();

          final SendMessage sendMessage =
              new SendMessage(chat.getId(), message).parseMode(ParseMode.Markdown);

          final InlineKeyboardButton[] buttons =
              modelHelp.generateInlineKeyBoardButton(chat, eventProcessor);

          if (buttons.length > 0) {
            sendMessage.replyMarkup(new InlineKeyboardMarkup(buttons));
          }

          return Optional.of(sendMessage);
        },
        threadPool);
  }
}
