package com.github.polpetta.mezzotre.telegram.command;

import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.google.inject.Inject;
import com.google.inject.Singleton;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.request.BaseRequest;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Named;

/**
 * This class has the goal of dispatching incoming {@link Update} events to the right {@link
 * Processor}, that will provide an adequate response.
 *
 * @author Davide Polonio
 * @since 1.0
 */
@Singleton
public class Router {

  private final Map<String, Processor> tgCommandProcessors;
  private final Executor threadPool;
  private final NotFoundFactory notFoundFactory;

  @Inject
  public Router(
      @Named("commandProcessor") Map<String, Processor> tgCommandProcessors,
      @Named("eventThreadPool") Executor threadPool,
      NotFoundFactory notFoundFactory) {
    this.tgCommandProcessors = tgCommandProcessors;
    this.threadPool = threadPool;
    this.notFoundFactory = notFoundFactory;
  }

  /**
   * Process the incoming {@link Update}. If no suitable {@link Processor} is able to process the
   * command, then {@link CommandNotFoundException} is used to signal this event.
   *
   * @param update the update coming from Telegram Servers
   * @return a {@link CompletableFuture} that is marked as failure and containing a {@link
   *     CommandNotFoundException} exception if no suitable {@link Processor} is found
   */
  public CompletableFuture<Optional<BaseRequest<?, ?>>> process(TgChat chat, Update update) {
    // This way exceptions are always under control
    return CompletableFuture.completedFuture(update)
        .thenComposeAsync(
            ignored ->
                /*
                Brief explanation of this chain:
                1 - Check if the message has a command in it (e.g. "/start hey!")
                  1.a - If it has, then search for the proper executor and go with it
                2 - If there is no command (e.g "example") then go to the chat context and retrieve the stage we
                    could be in (maybe we're continuing a chat from previous messages?)
                    2.a - If there's a context with a valid stage, then continue with it
                 */
                Optional.of(update.message().text().split(" "))
                    .filter(list -> list.length > 0)
                    .map(list -> list[0])
                    .filter(wannabeCommand -> wannabeCommand.startsWith("/"))
                    .or(() -> Optional.ofNullable(chat.getChatContext().getStage()))
                    .map(
                        command ->
                            tgCommandProcessors
                                .getOrDefault(command, notFoundFactory.create(command))
                                .process(chat, update))
                    // This should never happen
                    .orElse(CompletableFuture.failedFuture(new IllegalStateException())),
            threadPool);
  }
}
