package com.github.polpetta.mezzotre.telegram.command;

import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.telegram.callbackquery.Field;
import com.github.polpetta.mezzotre.telegram.callbackquery.Value;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.google.inject.Singleton;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.model.request.ParseMode;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import java.util.Collections;
import java.util.Locale;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Inject;
import javax.inject.Named;
import org.slf4j.Logger;

/**
 * This {@link Processor} has the goal to greet a user that typed {@code /start} to the bot.
 *
 * @author Davide Polonio
 * @since 1.0
 */
@Singleton
class Start implements Processor {

  private static final String TRIGGERING_STAGING_NAME = "/start";

  private final Executor threadPool;
  private final Logger log;
  private final UUIDGenerator uuidGenerator;
  private final String applicationName;
  private final ChatUtil chatUtil;

  private final TemplateContentGenerator templateContentGenerator;

  @Inject
  public Start(
      TemplateContentGenerator templateContentGenerator,
      @Named("eventThreadPool") Executor threadPool,
      Logger log,
      UUIDGenerator uuidGenerator,
      @Named("applicationName") String applicationName,
      ChatUtil chatUtil) {
    this.templateContentGenerator = templateContentGenerator;
    this.threadPool = threadPool;
    this.log = log;
    this.uuidGenerator = uuidGenerator;
    this.applicationName = applicationName;
    this.chatUtil = chatUtil;
  }

  @Override
  public Set<String> getTriggerKeywords() {
    return Set.of(TRIGGERING_STAGING_NAME);
  }

  @Override
  public CompletableFuture<Optional<BaseRequest<?, ?>>> process(TgChat chat, Update update) {
    return CompletableFuture.supplyAsync(
        () -> {
          /*
          Steps:
            1 - Load the template and generate the message with the defined locale
            2 - Update the chat context
            3 - Reply to Telegram
           */
          final String message =
              templateContentGenerator.mergeTemplate(
                  velocityContext -> {
                    velocityContext.put("firstName", update.message().chat().firstName());
                    // FIXME add some very cool markdown formatter instead of concatenating stuff
                    // this way
                    velocityContext.put("programName", "_" + applicationName + "_");
                  },
                  chat.getLocale(),
                  "template/telegram/start.vm");
          log.trace("Start command - message to send back: " + message);

          chatUtil.updateChatContext(chat, TRIGGERING_STAGING_NAME, 0, Collections.emptyMap());

          // To get the messageId we should send the message first, then save it in the database!
          final String groupId = uuidGenerator.generateAsString();
          final CallbackQueryContext switchToEnglish =
              new CallbackQueryContext(
                  uuidGenerator.generateAsString(),
                  groupId,
                  new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                      .withEvent(
                          com.github.polpetta.mezzotre.telegram.callbackquery.SelectLanguageTutorial
                              .EVENT_NAME)
                      .withTelegramChatId(update.message().chat().id())
                      .withAdditionalProperty(
                          Field.SelectLanguageTutorial.NewLanguage.getName(),
                          Value.SelectLanguageTutorial.English.getLocale())
                      .build());

          final CallbackQueryContext switchToItalian =
              new CallbackQueryContext(
                  uuidGenerator.generateAsString(),
                  groupId,
                  new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                      .withEvent(
                          com.github.polpetta.mezzotre.telegram.callbackquery.SelectLanguageTutorial
                              .EVENT_NAME)
                      .withTelegramChatId(update.message().chat().id())
                      .withAdditionalProperty(
                          Field.SelectLanguageTutorial.NewLanguage.getName(),
                          Value.SelectLanguageTutorial.Italian.getLocale())
                      .build());

          final String englishButton =
              templateContentGenerator.getString(Locale.US, "changeLanguage.english");
          final String italianButton =
              templateContentGenerator.getString(Locale.ITALY, "changeLanguage.italian");

          final SendMessage messageToSend =
              new SendMessage(chat.getId(), message)
                  .parseMode(ParseMode.Markdown)
                  .replyMarkup(
                      new InlineKeyboardMarkup(
                          new InlineKeyboardButton(englishButton)
                              .callbackData(switchToEnglish.getId()),
                          new InlineKeyboardButton(italianButton)
                              .callbackData(switchToItalian.getId())));

          switchToEnglish.save();
          switchToItalian.save();

          return Optional.of(messageToSend);
        },
        threadPool);
  }
}
