package com.github.polpetta.mezzotre.telegram.model;

import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.telegram.callbackquery.Field;
import com.github.polpetta.mezzotre.telegram.command.Processor;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import java.util.Map;
import java.util.stream.Collectors;
import javax.inject.Inject;
import org.apache.commons.lang3.tuple.Pair;

/**
 * This class provides the model for the Help message output. It is the business logic behind the
 * message, and a way to keep the codebase DRY.
 *
 * @author Davide Polonio
 * @since 1.0
 * @see com.github.polpetta.mezzotre.telegram.command.Help for the command help
 * @see com.github.polpetta.mezzotre.telegram.callbackquery.ShowHelp for the event help
 */
public class Help {

  private final TemplateContentGenerator templateContentGenerator;
  private final UUIDGenerator uuidGenerator;

  @Inject
  public Help(TemplateContentGenerator templateContentGenerator, UUIDGenerator uuidGenerator) {
    this.templateContentGenerator = templateContentGenerator;
    this.uuidGenerator = uuidGenerator;
  }

  /**
   * Generates the message that will be sent back to the user. This takes the given {@link
   * Processor} and formats them accordingly to the chat locale
   *
   * @param chat the {@link TgChat} conversation
   * @param tgCommandProcessors a {@link Map} of all the {@link Processor} that will be printed in
   *     the help message
   * @return a {@link String} localized ready to be sent to the user
   */
  public String getMessage(TgChat chat, Map<String, Processor> tgCommandProcessors) {
    return templateContentGenerator.mergeTemplate(
        velocityContext -> {
          velocityContext.put(
              "commands",
              tgCommandProcessors.values().stream()
                  .distinct()
                  .map(
                      p ->
                          Pair.of(
                              p.getTriggerKeywords().stream().sorted().collect(Collectors.toList()),
                              p.getLocaleDescriptionKeyword()))
                  .collect(Collectors.toList()));
        },
        chat.getLocale(),
        "template/telegram/help.vm");
  }

  /**
   * Generates {@link InlineKeyboardButton} to be returned to the user to give them the possibility
   * to interact with them via events rather than commands.
   *
   * @param chat the current {@link TgChat} conversation
   * @param eventProcessors a {@link Map} of {@link
   *     com.github.polpetta.mezzotre.telegram.callbackquery.Processor} that are currently used to
   *     process all events. Note that only the one the user can interact with will be added as
   *     buttons
   * @return an array of {@link InlineKeyboardButton}
   */
  public InlineKeyboardButton[] generateInlineKeyBoardButton(
      TgChat chat,
      Map<String, com.github.polpetta.mezzotre.telegram.callbackquery.Processor> eventProcessors) {
    final String callBackGroupId = uuidGenerator.generateAsString();
    return eventProcessors.values().stream()
        .filter(
            com.github.polpetta.mezzotre.telegram.callbackquery.Processor
                ::canBeDirectlyInvokedByTheUser)
        .filter(e -> e.getPrettyPrintLocaleKeyName().isPresent())
        .map(
            eventProcessor -> {
              final CallbackQueryContext callbackQueryContext =
                  new CallbackQueryContext(
                      uuidGenerator.generateAsString(),
                      callBackGroupId,
                      new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                          .withEvent(eventProcessor.getEventName())
                          .withTelegramChatId(chat.getId())
                          .withAdditionalProperty(
                              Field.ShowHelp.InvokedFromHelpMessage.getName(), true)
                          .build());
              callbackQueryContext.save();

              return new InlineKeyboardButton(
                      templateContentGenerator.getString(
                          chat.getLocale(), eventProcessor.getPrettyPrintLocaleKeyName().get()))
                  .callbackData(callbackQueryContext.getId());
            })
        .toArray(InlineKeyboardButton[]::new);
  }
}
