package com.github.polpetta.mezzotre.orm;

import static org.junit.jupiter.api.Assertions.*;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import io.ebean.Database;
import java.time.Duration;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import org.apache.commons.lang3.tuple.Pair;
import org.junit.jupiter.api.*;
import org.slf4j.LoggerFactory;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Tag("velocity")
@Testcontainers
class BatchBeanCleanerServiceIntegrationTest {

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private Database database;
  private BatchBeanCleanerService batchBeanCleanerService;

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));
    batchBeanCleanerService =
        new BatchBeanCleanerService(
            LoggerFactory.getLogger(BatchBeanCleanerService.class),
            Pair.of(0, TimeUnit.MILLISECONDS));
  }

  @AfterEach
  void tearDown() throws Exception {
    if (batchBeanCleanerService != null) {
      batchBeanCleanerService.stopAsync().awaitTerminated(Duration.ofSeconds(10));
    }
  }

  @Test
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  void shouldRemoveEntryFromDatabase() throws Exception {
    final CallbackQueryContext callbackQueryContext =
        new CallbackQueryContext("1234", "4567", new CallbackQueryMetadata());
    callbackQueryContext.save();

    batchBeanCleanerService.startAsync();
    batchBeanCleanerService.awaitRunning();

    final CompletableFuture<Integer> integerCompletableFuture =
        batchBeanCleanerService.removeAsync("1234", new QCallbackQueryContext().id);

    final Integer gotDeletion = integerCompletableFuture.get();

    assertEquals(1, gotDeletion);
  }

  @Test
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  void shouldRemoveMultipleEntriesFromDatabase() throws Exception {
    final CallbackQueryContext callbackQueryContext =
        new CallbackQueryContext("1234", "4567", new CallbackQueryMetadata());
    callbackQueryContext.save();
    final CallbackQueryContext callbackQueryContext2 =
        new CallbackQueryContext("4321", "4567", new CallbackQueryMetadata());
    callbackQueryContext2.save();

    batchBeanCleanerService.startAsync();
    batchBeanCleanerService.awaitRunning();

    final CompletableFuture<Integer> integerCompletableFuture =
        batchBeanCleanerService.removeAsync("4567", new QCallbackQueryContext().entryGroup);
    final CompletableFuture<Integer> integerCompletableFuture2 =
        batchBeanCleanerService.removeAsync("4567", new QCallbackQueryContext().entryGroup);

    final Integer gotDeletion1 = integerCompletableFuture.get();
    final Integer gotDeletion2 = integerCompletableFuture2.get();

    assertEquals(2, gotDeletion1);
    assertEquals(0, gotDeletion2);
  }
}
