package com.github.polpetta.mezzotre.orm.model;

import static org.junit.jupiter.api.Assertions.*;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.mezzotre.util.Clock;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import io.ebean.Database;
import io.ebean.SqlRow;
import java.sql.Timestamp;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Testcontainers
class CallbackQueryDIContextIntegrationTest {

  private static ObjectMapper objectMapper;
  private static UUIDGenerator uuidGenerator;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private Database database;

  @BeforeAll
  static void beforeAll() {
    objectMapper = new ObjectMapper();
    uuidGenerator = new UUIDGenerator();
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));
  }

  @Test
  void shouldInsertEntryIntoDatabase() throws Exception {
    final String uuidContext = uuidGenerator.generateAsString();
    final String uuidGroup = uuidGenerator.generateAsString();
    final CallbackQueryMetadata callbackQueryMetadata =
        new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
            .withEvent("eventExample")
            .withTelegramChatId(1234L)
            .withMessageId(4242L)
            .build();

    final CallbackQueryContext callbackQueryContext =
        new CallbackQueryContext(uuidContext, uuidGroup, callbackQueryMetadata);
    callbackQueryContext.save();

    final String query = "select * from callback_query_context where id = ?";
    final SqlRow savedCallbackQueryContext =
        database.sqlQuery(query).setParameter(uuidContext).findOne();

    assertNotNull(savedCallbackQueryContext);
    assertEquals(uuidGroup, savedCallbackQueryContext.getString("entry_group"));
    assertEquals(uuidContext, savedCallbackQueryContext.getString("id"));
    assertEquals(
        objectMapper.writeValueAsString(callbackQueryMetadata),
        savedCallbackQueryContext.getString("fields"));
  }

  @Test
  void shouldRetrieveEntryInDatabase() throws Exception {
    final Timestamp timestampFromUnixEpoch = Clock.getTimestampFromUnixEpoch(1L);
    final String uuidId = uuidGenerator.generateAsString();
    final String uuidEntryGroup = uuidGenerator.generateAsString();
    final CallbackQueryMetadata callbackQueryMetadata =
        new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
            .withEvent("eventExample")
            .withTelegramChatId(12324L)
            .withMessageId(42342L)
            .build();
    final String insertQuery = "insert into callback_query_context values (?, ?, ?::json, ?, ?)";
    final int affectedRows =
        database
            .sqlUpdate(insertQuery)
            .setParameter(uuidId)
            .setParameter(uuidEntryGroup)
            .setParameter(objectMapper.writeValueAsString(callbackQueryMetadata))
            .setParameter(timestampFromUnixEpoch)
            .setParameter(timestampFromUnixEpoch)
            .execute();

    assertEquals(1, affectedRows);

    final CallbackQueryContext got = new QCallbackQueryContext().id.eq(uuidId).findOne();
    assertNotNull(got);
    final CallbackQueryMetadata gotFields = got.getFields();
    assertNotNull(gotFields);
    assertEquals("eventExample", gotFields.getEvent());
    assertEquals(12324L, Double.valueOf(gotFields.getTelegramChatId()).longValue());
    assertEquals(42342L, Double.valueOf(gotFields.getMessageId()).longValue());
  }
}
