package com.github.polpetta.mezzotre.orm.telegram;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.util.Clock;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import io.ebean.Database;
import java.util.Optional;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Testcontainers
class ChatUtilIntegrationTest {

  private static Gson gson;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private Database database;
  private Clock fakeClock;
  private ChatUtil chatUtil;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));

    fakeClock = mock(Clock.class);
    chatUtil = new ChatUtil(fakeClock);
  }

  @Test
  void shouldExtractChatFromCallBackQueryContext() {
    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/help\"\n"
                + "}\n"
                + "}",
            Update.class);
    final CallbackQueryContext callbackQueryContext =
        new CallbackQueryContext(
            "123",
            "456",
            new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                .withTelegramChatId(69420L)
                .build());

    final TgChat expectedChat = new TgChat(69420L, new ChatContext());
    expectedChat.save();

    final Optional<TgChat> tgChatOptional =
        chatUtil.extractChat(
            callbackQueryContext, null); // null just for test purposes, not usually expected

    assertEquals(expectedChat, tgChatOptional.get());
  }

  @Test
  void shouldExtractChatFromUpdateIfCallbackQueryContextIsEmpty() {
    final Update update =
        gson.fromJson(
            "{\n"
                + "  \"update_id\": 158712614,\n"
                + "  \"callback_query\": {\n"
                + "    \"id\": \"20496049451114620\",\n"
                + "    \"from\": {\n"
                + "      \"id\": 1111111,\n"
                + "      \"is_bot\": false,\n"
                + "      \"first_name\": \"Test Firstname\",\n"
                + "      \"last_name\": \"Test Lastname\",\n"
                + "      \"username\": \"Testusername\",\n"
                + "      \"language_code\": \"en\"\n"
                + "    },\n"
                + "    \"message\": {\n"
                + "      \"message_id\": 69420,\n"
                + "      \"from\": {\n"
                + "        \"id\": 244745330,\n"
                + "        \"is_bot\": true,\n"
                + "        \"first_name\": \"Dev - DavideBot\",\n"
                + "        \"username\": \"devdavidebot\"\n"
                + "      },\n"
                + "      \"date\": 1681218838,\n"
                + "      \"chat\": {\n"
                + "        \"id\": 1111111,\n"
                + "        \"type\": \"private\",\n"
                + "        \"username\": \"Testusername\",\n"
                + "        \"first_name\": \"Test Firstname\",\n"
                + "        \"last_name\": \"Test Lastname\"\n"
                + "      },\n"
                + "      \"text\": \"Hello xxxxxx! \uD83D\uDC4B\\n"
                + "\\n"
                + "This is Mezzotre, a simple bot focused on DnD content management! Please start"
                + " by choosing a language down below \uD83D\uDC47\",\n"
                + "      \"entities\": [\n"
                + "        {\n"
                + "          \"type\": \"bold\",\n"
                + "          \"offset\": 0,\n"
                + "          \"length\": 16\n"
                + "        },\n"
                + "        {\n"
                + "          \"type\": \"italic\",\n"
                + "          \"offset\": 26,\n"
                + "          \"length\": 8\n"
                + "        }\n"
                + "      ],\n"
                + "      \"reply_markup\": {\n"
                + "        \"inline_keyboard\": [\n"
                + "          [\n"
                + "            {\n"
                + "              \"text\": \"English\",\n"
                + "              \"callback_data\": \"9a64be11-d086-4bd9-859f-720c43dedcb5\"\n"
                + "            },\n"
                + "            {\n"
                + "              \"text\": \"Italian\",\n"
                + "              \"callback_data\": \"8768d660-f05f-4f4b-bda5-3451ab573d56\"\n"
                + "            }\n"
                + "          ]\n"
                + "        ]\n"
                + "      }\n"
                + "    }\n"
                + "  }\n"
                + "}",
            Update.class);
    final CallbackQueryContext callbackQueryContext =
        new CallbackQueryContext("123", "456", new CallbackQueryMetadata());

    final TgChat expectedChat = new TgChat(69420L, new ChatContext());
    expectedChat.save();

    final Optional<TgChat> tgChatOptional =
        chatUtil.extractChat(
            callbackQueryContext, update); // null just for test purposes, not usually expected

    assertEquals(expectedChat, tgChatOptional.get());
  }
}
