package com.github.polpetta.mezzotre.telegram.callbackquery;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.BatchBeanCleanerService;
import com.github.polpetta.mezzotre.orm.CallbackQueryContextCleaner;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.telegram.model.Help;
import com.github.polpetta.mezzotre.util.Clock;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.EditMessageText;
import com.pengrad.telegrambot.request.SendMessage;
import io.ebean.Database;
import io.ebean.typequery.PString;
import io.ebean.typequery.TQRootBean;
import io.vavr.Tuple3;
import java.time.Duration;
import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Stream;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.velocity.app.VelocityEngine;
import org.junit.jupiter.api.*;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Tag("velocity")
@Testcontainers
class ShowHelpIntegrationTest {

  private static Gson gson;
  private static Logger testLog;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private Database database;
  private VelocityEngine velocityEngine;
  private UUIDGenerator fakeUUIDGenerator;
  private Clock fakeClock;
  private BatchBeanCleanerService batchBeanCleanerService;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
    testLog = LoggerFactory.getLogger(ShowHelpIntegrationTest.class);
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));
    velocityEngine = Loader.defaultVelocityEngine();

    final Logger log = LoggerFactory.getLogger(ShowHelp.class);
    fakeUUIDGenerator = mock(UUIDGenerator.class);
    fakeClock = mock(Clock.class);
  }

  @AfterEach
  void tearDown() throws Exception {
    if (batchBeanCleanerService != null) {
      batchBeanCleanerService.stopAsync().awaitTerminated(Duration.ofSeconds(10));
    }
  }

  private ShowHelp generateShowHelpWithCustomCommands(
      Map<String, Processor> eventProcessors,
      Map<String, com.github.polpetta.mezzotre.telegram.command.Processor> commandProcessors)
      throws TimeoutException {

    batchBeanCleanerService =
        new BatchBeanCleanerService(
            LoggerFactory.getLogger(CallbackQueryContextCleaner.class),
            Pair.of(0, TimeUnit.MILLISECONDS));
    batchBeanCleanerService.startAsync().awaitRunning(Duration.ofSeconds(10));

    return new ShowHelp(
        Executors.newSingleThreadExecutor(),
        new Help(
            new TemplateContentGenerator(new LocalizedMessageFactory(velocityEngine)),
            fakeUUIDGenerator),
        commandProcessors,
        eventProcessors,
        // We need to implement a service that is sync - no thread creation!
        new ChatUtil(fakeClock),
        new CallbackQueryContextCleaner(
            batchBeanCleanerService, LoggerFactory.getLogger(CallbackQueryContextCleaner.class)));
  }

  public static Stream<Arguments> getMessageIdOrNot() {
    return Stream.of(
        Arguments.of(
            gson.fromJson(
                "{\n"
                    + "  \"update_id\": 158712614,\n"
                    + "  \"callback_query\": {\n"
                    + "    \"id\": \"20496049451114620\",\n"
                    + "    \"from\": {\n"
                    + "      \"id\": 1111111,\n"
                    + "      \"is_bot\": false,\n"
                    + "      \"first_name\": \"Test Firstname\",\n"
                    + "      \"last_name\": \"Test Lastname\",\n"
                    + "      \"username\": \"Testusername\",\n"
                    + "      \"language_code\": \"en\"\n"
                    + "    },\n"
                    + "    \"message\": {\n"
                    + "      \"message_id\": 2723,\n"
                    + "      \"from\": {\n"
                    + "        \"id\": 244745330,\n"
                    + "        \"is_bot\": true,\n"
                    + "        \"first_name\": \"Dev - DavideBot\",\n"
                    + "        \"username\": \"devdavidebot\"\n"
                    + "      },\n"
                    + "      \"date\": 1681218838,\n"
                    + "      \"chat\": {\n"
                    + "        \"id\": 1111111,\n"
                    + "        \"type\": \"private\",\n"
                    + "        \"username\": \"Testusername\",\n"
                    + "        \"first_name\": \"Test Firstname\",\n"
                    + "        \"last_name\": \"Test Lastname\"\n"
                    + "      },\n"
                    + "      \"text\": \"a message\",\n"
                    + "      \"reply_markup\": {\n"
                    + "        \"inline_keyboard\": [\n"
                    + "          [\n"
                    + "            {\n"
                    + "              \"text\": \"English\",\n"
                    + "              \"callback_data\": \"9a64be11-d086-4bd9-859f-720c43dedcb5\"\n"
                    + "            },\n"
                    + "            {\n"
                    + "              \"text\": \"Italian\",\n"
                    + "              \"callback_data\": \"8768d660-f05f-4f4b-bda5-3451ab573d56\"\n"
                    + "            }\n"
                    + "          ]\n"
                    + "        ]\n"
                    + "      }\n"
                    + "    }\n"
                    + "  }\n"
                    + "}",
                Update.class),
            EditMessageText.class),
        Arguments.of(
            gson.fromJson(
                "{\n"
                    + "  \"update_id\": 158712614,\n"
                    + "  \"callback_query\": {\n"
                    + "    \"id\": \"20496049451114620\",\n"
                    + "    \"from\": {\n"
                    + "      \"id\": 1111111,\n"
                    + "      \"is_bot\": false,\n"
                    + "      \"first_name\": \"Test Firstname\",\n"
                    + "      \"last_name\": \"Test Lastname\",\n"
                    + "      \"username\": \"Testusername\",\n"
                    + "      \"language_code\": \"en\"\n"
                    + "    }\n"
                    + "  }\n"
                    + "}",
                Update.class),
            SendMessage.class));
  }

  @ParameterizedTest
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  @MethodSource("getMessageIdOrNot")
  void shouldShowHelpMessageWithButtons(Update update, Class<?> typeOfMessage) throws Exception {
    // we have 2 events + group uuid
    when(fakeUUIDGenerator.generateAsString())
        .thenReturn("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5")
        .thenReturn("16507fbd-9f28-48a8-9de1-3ea1c943af67")
        .thenReturn("0b0ac18e-f621-484e-aa8d-9b176be5b930");
    when(fakeClock.now()).thenReturn(42L);

    final HashMap<String, com.github.polpetta.mezzotre.telegram.command.Processor> commands =
        new HashMap<>();
    final com.github.polpetta.mezzotre.telegram.command.Processor dummy1 =
        new com.github.polpetta.mezzotre.telegram.command.Processor() {
          @Override
          public Set<String> getTriggerKeywords() {
            return Set.of("/a", "/b");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              TgChat chat, Update update) {
            return null;
          }

          @Override
          public String getLocaleDescriptionKeyword() {
            return "start.cmdDescription";
          }
        };
    final com.github.polpetta.mezzotre.telegram.command.Processor dummy2 =
        new com.github.polpetta.mezzotre.telegram.command.Processor() {
          @Override
          public Set<String> getTriggerKeywords() {
            return Set.of("/different");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              TgChat chat, Update update) {
            return null;
          }

          @Override
          public String getLocaleDescriptionKeyword() {
            return "help.cmdDescription";
          }
        };
    commands.put("/a", dummy1);
    commands.put("/b", dummy1);
    commands.put("/different", dummy2);

    final Map<String, Processor> events = new HashMap<>();

    final Processor dummyEvent1 =
        new Processor() {
          @Override
          public String getEventName() {
            return "exampleEvent";
          }

          @Override
          public boolean canBeDirectlyInvokedByTheUser() {
            return true;
          }

          @Override
          public Optional<String> getPrettyPrintLocaleKeyName() {
            return Optional.of("changeLanguage.inlineKeyboardButtonName");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              CallbackQueryContext callbackQueryContext, Update update) {
            return null;
          }
        };

    final Processor dummyEvent2 =
        new com.github.polpetta.mezzotre.telegram.callbackquery.Processor() {
          @Override
          public String getEventName() {
            return "secondExampleEvent";
          }

          @Override
          public boolean canBeDirectlyInvokedByTheUser() {
            return true;
          }

          @Override
          public Optional<String> getPrettyPrintLocaleKeyName() {
            return Optional.of("selectLanguageTutorial.inlineKeyboardButtonName");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              CallbackQueryContext callbackQueryContext, Update update) {
            return null;
          }
        };

    events.put(dummyEvent1.getEventName(), dummyEvent1);
    events.put(dummyEvent2.getEventName(), dummyEvent2);

    final ShowHelp showHelp = generateShowHelpWithCustomCommands(events, commands);
    final TgChat tgChat = new TgChat(1111111L, new ChatContext());
    tgChat.save();
    final CallbackQueryContext callbackQueryContext =
        new CallbackQueryContext(
            "1234",
            "5678",
            new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                .withTelegramChatId(1111111L)
                .build());
    callbackQueryContext.save();

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotResultFuture =
        showHelp.process(callbackQueryContext, update);

    final Optional<BaseRequest<?, ?>> baseRequestOptional = gotResultFuture.get();
    final BaseRequest<?, ?> gotResponse = baseRequestOptional.get();
    assertInstanceOf(typeOfMessage, gotResponse);
    final String message = (String) gotResponse.getParameters().get("text");
    assertEquals(
        "Here is a list of what I can do:\n"
            + "\n"
            + "- /a /b: Trigger this very bot\n"
            + "- /different: Print the help message\n"
            + "\n"
            + "You can do the same operations you'd do with the commands aforementioned by"
            + " selecting the corresponding button below \uD83D\uDC47",
        message);
    final InlineKeyboardButton[][] keyboard =
        ((InlineKeyboardMarkup)
                gotResponse
                    .getParameters()
                    .getOrDefault("reply_markup", new InlineKeyboardMarkup()))
            .inlineKeyboard();

    final List<InlineKeyboardButton> keyboardButtons =
        Stream.of(keyboard).flatMap(Stream::of).toList();

    assertEquals(2, keyboardButtons.size());

    assertFalse(keyboardButtons.get(0).callbackData().isBlank());
    assertFalse(keyboardButtons.get(1).callbackData().isBlank());

    assertTrue(
        keyboardButtons.stream()
            .map(InlineKeyboardButton::text)
            .anyMatch("Select language"::equals));

    assertTrue(
        keyboardButtons.stream()
            .map(InlineKeyboardButton::text)
            .anyMatch("Change language"::equals));

    final TgChat gotChat = new QTgChat().id.eq(1111111L).findOne();
    assertNotNull(gotChat);
    assertTrue(gotChat.getHasHelpBeenShown());
    final ChatContext gotChatChatContext = gotChat.getChatContext();
    assertEquals(0, gotChatChatContext.getStep());
    assertEquals("showHelp", gotChatChatContext.getStage());
    assertEquals(42, gotChatChatContext.getPreviousMessageUnixTimestampInSeconds());
  }

  @Test
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  void shouldCleanUpCallbackQueryContextAfterRun() throws Exception {
    when(fakeClock.now()).thenReturn(42L);

    final ShowHelp showHelp =
        generateShowHelpWithCustomCommands(Collections.emptyMap(), Collections.emptyMap());
    final TgChat tgChat = new TgChat(1111111L, new ChatContext());
    tgChat.save();
    final Update update =
        gson.fromJson(
            "{\n"
                + "  \"update_id\": 158712614,\n"
                + "  \"callback_query\": {\n"
                + "    \"id\": \"20496049451114620\",\n"
                + "    \"from\": {\n"
                + "      \"id\": 1111111,\n"
                + "      \"is_bot\": false,\n"
                + "      \"first_name\": \"Test Firstname\",\n"
                + "      \"last_name\": \"Test Lastname\",\n"
                + "      \"username\": \"Testusername\",\n"
                + "      \"language_code\": \"en\"\n"
                + "    },\n"
                + "    \"message\": {\n"
                + "      \"message_id\": 2723,\n"
                + "      \"from\": {\n"
                + "        \"id\": 244745330,\n"
                + "        \"is_bot\": true,\n"
                + "        \"first_name\": \"Dev - DavideBot\",\n"
                + "        \"username\": \"devdavidebot\"\n"
                + "      },\n"
                + "      \"date\": 1681218838,\n"
                + "      \"chat\": {\n"
                + "        \"id\": 1111111,\n"
                + "        \"type\": \"private\",\n"
                + "        \"username\": \"Testusername\",\n"
                + "        \"first_name\": \"Test Firstname\",\n"
                + "        \"last_name\": \"Test Lastname\"\n"
                + "      },\n"
                + "      \"text\": \"Hello xxxxxx! \uD83D\uDC4B\\n"
                + "\\n"
                + "This is Mezzotre, a simple bot focused on DnD content management! Please start"
                + " by choosing a language down below \uD83D\uDC47\",\n"
                + "      \"entities\": [\n"
                + "        {\n"
                + "          \"type\": \"bold\",\n"
                + "          \"offset\": 0,\n"
                + "          \"length\": 16\n"
                + "        },\n"
                + "        {\n"
                + "          \"type\": \"italic\",\n"
                + "          \"offset\": 26,\n"
                + "          \"length\": 8\n"
                + "        }\n"
                + "      ]\n"
                + "    }\n"
                + "  }\n"
                + "}",
            Update.class);
    final CallbackQueryContext callbackQueryContext =
        new CallbackQueryContext(
            "1234",
            "5678",
            new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                .withTelegramChatId(1111111L)
                .build());
    callbackQueryContext.save();
    // Create another CallbackQuery context of the same group
    final CallbackQueryContext callbackQueryContext2 =
        new CallbackQueryContext(
            "7891",
            "5678",
            new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                .withTelegramChatId(1111111L)
                .build());
    callbackQueryContext2.save();

    // I know I know, a future containing a future...but it is for testing purposes, otherwise we'd
    // have to do some while/sleep thread hack which I find horrible to do
    final CompletableFuture<
            Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>>
        callBackFuture = new CompletableFuture<>();
    batchBeanCleanerService.addListener(callBackFuture::complete);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotResultFuture =
        showHelp.process(callbackQueryContext, update);

    final Optional<BaseRequest<?, ?>> baseRequestOptional = gotResultFuture.get();
    assertDoesNotThrow(baseRequestOptional::get);

    final Tuple3<String, PString<? extends TQRootBean<?, ?>>, CompletableFuture<Integer>>
        deletionCallback = callBackFuture.get();

    assertEquals("5678", deletionCallback._1());
    assertEquals(2, deletionCallback._3().get());
    assertEquals(
        0,
        new QCallbackQueryContext().findCount(),
        "The CallbackQuery context group has not been cleaned properly!");
  }
}
