package com.github.polpetta.mezzotre.telegram.callbackquery;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.mockito.Mockito.*;

import com.github.polpetta.mezzotre.orm.CallbackQueryContextCleaner;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.telegram.model.Help;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.request.BaseRequest;
import java.util.Collections;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.parallel.Execution;
import org.junit.jupiter.api.parallel.ExecutionMode;

@Execution(ExecutionMode.CONCURRENT)
class ShowHelpTest {

  private static Gson gson;
  private ShowHelp showHelp;
  private Help fakeModelHelp;
  private ChatUtil fakeChatUtil;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() {

    fakeModelHelp = mock(Help.class);
    fakeChatUtil = mock(ChatUtil.class);
    final CallbackQueryContextCleaner fakeCallbackQueryContextCleaner =
        mock(CallbackQueryContextCleaner.class);

    showHelp =
        new ShowHelp(
            Executors.newSingleThreadExecutor(),
            fakeModelHelp,
            Collections.emptyMap(),
            Collections.emptyMap(),
            fakeChatUtil,
            fakeCallbackQueryContextCleaner);
  }

  @Test
  void shouldUpdateChatContext() throws Exception {
    final CallbackQueryContext fakeCallbackQueryContext = mock(CallbackQueryContext.class);
    final Update update =
        gson.fromJson(
            "{\n"
                + "  \"update_id\": 158712614,\n"
                + "  \"callback_query\": {\n"
                + "    \"id\": \"20496049451114620\",\n"
                + "    \"from\": {\n"
                + "      \"id\": 1111111,\n"
                + "      \"is_bot\": false,\n"
                + "      \"first_name\": \"Test Firstname\",\n"
                + "      \"last_name\": \"Test Lastname\",\n"
                + "      \"username\": \"Testusername\",\n"
                + "      \"language_code\": \"en\"\n"
                + "    },\n"
                + "    \"message\": {\n"
                + "      \"message_id\": 2723,\n"
                + "      \"from\": {\n"
                + "        \"id\": 244745330,\n"
                + "        \"is_bot\": true,\n"
                + "        \"first_name\": \"Dev - DavideBot\",\n"
                + "        \"username\": \"devdavidebot\"\n"
                + "      },\n"
                + "      \"date\": 1681218838,\n"
                + "      \"chat\": {\n"
                + "        \"id\": 1111111,\n"
                + "        \"type\": \"private\",\n"
                + "        \"username\": \"Testusername\",\n"
                + "        \"first_name\": \"Test Firstname\",\n"
                + "        \"last_name\": \"Test Lastname\"\n"
                + "      },\n"
                + "      \"text\": \"Hello xxxxxx! \uD83D\uDC4B\\n"
                + "\\n"
                + "This is Mezzotre, a simple bot focused on DnD content management! Please start"
                + " by choosing a language down below \uD83D\uDC47\",\n"
                + "      \"entities\": [\n"
                + "        {\n"
                + "          \"type\": \"bold\",\n"
                + "          \"offset\": 0,\n"
                + "          \"length\": 16\n"
                + "        },\n"
                + "        {\n"
                + "          \"type\": \"italic\",\n"
                + "          \"offset\": 26,\n"
                + "          \"length\": 8\n"
                + "        }\n"
                + "      ],\n"
                + "      \"reply_markup\": {\n"
                + "        \"inline_keyboard\": [\n"
                + "          [\n"
                + "            {\n"
                + "              \"text\": \"English\",\n"
                + "              \"callback_data\": \"9a64be11-d086-4bd9-859f-720c43dedcb5\"\n"
                + "            },\n"
                + "            {\n"
                + "              \"text\": \"Italian\",\n"
                + "              \"callback_data\": \"8768d660-f05f-4f4b-bda5-3451ab573d56\"\n"
                + "            }\n"
                + "          ]\n"
                + "        ]\n"
                + "      }\n"
                + "    }\n"
                + "  }\n"
                + "}",
            Update.class);
    final TgChat fakeTgChat = mock(TgChat.class);
    when(fakeTgChat.getId()).thenReturn(1111111L);
    when(fakeTgChat.getChatContext()).thenReturn(new ChatContext());
    when(fakeModelHelp.getMessage(eq(fakeTgChat), any())).thenReturn("doesn't matter");
    when(fakeModelHelp.generateInlineKeyBoardButton(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn(new InlineKeyboardButton[] {});
    when(fakeChatUtil.extractChat(eq(fakeCallbackQueryContext), eq(update)))
        .thenReturn(Optional.of(fakeTgChat));

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotResponseFuture =
        showHelp.process(fakeCallbackQueryContext, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = gotResponseFuture.get();

    verify(fakeChatUtil, times(1))
        .updateChatContext(eq(fakeTgChat), eq("showHelp"), eq(0), eq(Collections.emptyMap()));
  }

  @Test
  void shouldAddInlineKeyboardIfButtonsAreReturned() throws Exception {
    final CallbackQueryContext fakeCallbackQueryContext = mock(CallbackQueryContext.class);
    final Update update =
        gson.fromJson(
            "{\n"
                + "  \"update_id\": 158712614,\n"
                + "  \"callback_query\": {\n"
                + "    \"id\": \"20496049451114620\",\n"
                + "    \"from\": {\n"
                + "      \"id\": 1111111,\n"
                + "      \"is_bot\": false,\n"
                + "      \"first_name\": \"Test Firstname\",\n"
                + "      \"last_name\": \"Test Lastname\",\n"
                + "      \"username\": \"Testusername\",\n"
                + "      \"language_code\": \"en\"\n"
                + "    },\n"
                + "    \"message\": {\n"
                + "      \"message_id\": 2723,\n"
                + "      \"from\": {\n"
                + "        \"id\": 244745330,\n"
                + "        \"is_bot\": true,\n"
                + "        \"first_name\": \"Dev - DavideBot\",\n"
                + "        \"username\": \"devdavidebot\"\n"
                + "      },\n"
                + "      \"date\": 1681218838,\n"
                + "      \"chat\": {\n"
                + "        \"id\": 1111111,\n"
                + "        \"type\": \"private\",\n"
                + "        \"username\": \"Testusername\",\n"
                + "        \"first_name\": \"Test Firstname\",\n"
                + "        \"last_name\": \"Test Lastname\"\n"
                + "      },\n"
                + "      \"text\": \"Hello xxxxxx! \uD83D\uDC4B\\n"
                + "\\n"
                + "This is Mezzotre, a simple bot focused on DnD content management! Please start"
                + " by choosing a language down below \uD83D\uDC47\",\n"
                + "      \"entities\": [\n"
                + "        {\n"
                + "          \"type\": \"bold\",\n"
                + "          \"offset\": 0,\n"
                + "          \"length\": 16\n"
                + "        },\n"
                + "        {\n"
                + "          \"type\": \"italic\",\n"
                + "          \"offset\": 26,\n"
                + "          \"length\": 8\n"
                + "        }\n"
                + "      ],\n"
                + "      \"reply_markup\": {\n"
                + "        \"inline_keyboard\": [\n"
                + "          [\n"
                + "            {\n"
                + "              \"text\": \"English\",\n"
                + "              \"callback_data\": \"9a64be11-d086-4bd9-859f-720c43dedcb5\"\n"
                + "            },\n"
                + "            {\n"
                + "              \"text\": \"Italian\",\n"
                + "              \"callback_data\": \"8768d660-f05f-4f4b-bda5-3451ab573d56\"\n"
                + "            }\n"
                + "          ]\n"
                + "        ]\n"
                + "      }\n"
                + "    }\n"
                + "  }\n"
                + "}",
            Update.class);
    final TgChat fakeTgChat = mock(TgChat.class);
    when(fakeTgChat.getId()).thenReturn(1111111L);
    when(fakeTgChat.getChatContext()).thenReturn(new ChatContext());
    when(fakeModelHelp.getMessage(eq(fakeTgChat), any())).thenReturn("doesn't matter");
    when(fakeModelHelp.generateInlineKeyBoardButton(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn(new InlineKeyboardButton[] {new InlineKeyboardButton("example1")});
    when(fakeChatUtil.extractChat(eq(fakeCallbackQueryContext), eq(update)))
        .thenReturn(Optional.of(fakeTgChat));

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotResponseFuture =
        showHelp.process(fakeCallbackQueryContext, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = gotResponseFuture.get();
    final BaseRequest<?, ?> gotResponse = gotResponseOpt.get();

    final InlineKeyboardMarkup replyMarkup =
        (InlineKeyboardMarkup) gotResponse.getParameters().get("reply_markup");
    assertEquals(1, replyMarkup.inlineKeyboard()[0].length);
    assertEquals("example1", replyMarkup.inlineKeyboard()[0][0].text());
  }

  @Test
  void shouldNotAddAnyKeyboardIfThereAreNoButtons() throws Exception {
    final CallbackQueryContext fakeCallbackQueryContext = mock(CallbackQueryContext.class);
    final Update update =
        gson.fromJson(
            "{\n"
                + "  \"update_id\": 158712614,\n"
                + "  \"callback_query\": {\n"
                + "    \"id\": \"20496049451114620\",\n"
                + "    \"from\": {\n"
                + "      \"id\": 1111111,\n"
                + "      \"is_bot\": false,\n"
                + "      \"first_name\": \"Test Firstname\",\n"
                + "      \"last_name\": \"Test Lastname\",\n"
                + "      \"username\": \"Testusername\",\n"
                + "      \"language_code\": \"en\"\n"
                + "    },\n"
                + "    \"message\": {\n"
                + "      \"message_id\": 2723,\n"
                + "      \"from\": {\n"
                + "        \"id\": 244745330,\n"
                + "        \"is_bot\": true,\n"
                + "        \"first_name\": \"Dev - DavideBot\",\n"
                + "        \"username\": \"devdavidebot\"\n"
                + "      },\n"
                + "      \"date\": 1681218838,\n"
                + "      \"chat\": {\n"
                + "        \"id\": 1111111,\n"
                + "        \"type\": \"private\",\n"
                + "        \"username\": \"Testusername\",\n"
                + "        \"first_name\": \"Test Firstname\",\n"
                + "        \"last_name\": \"Test Lastname\"\n"
                + "      },\n"
                + "      \"text\": \"Hello xxxxxx! \uD83D\uDC4B\\n"
                + "\\n"
                + "This is Mezzotre, a simple bot focused on DnD content management! Please start"
                + " by choosing a language down below \uD83D\uDC47\",\n"
                + "      \"entities\": [\n"
                + "        {\n"
                + "          \"type\": \"bold\",\n"
                + "          \"offset\": 0,\n"
                + "          \"length\": 16\n"
                + "        },\n"
                + "        {\n"
                + "          \"type\": \"italic\",\n"
                + "          \"offset\": 26,\n"
                + "          \"length\": 8\n"
                + "        }\n"
                + "      ],\n"
                + "      \"reply_markup\": {\n"
                + "        \"inline_keyboard\": [\n"
                + "          [\n"
                + "            {\n"
                + "              \"text\": \"English\",\n"
                + "              \"callback_data\": \"9a64be11-d086-4bd9-859f-720c43dedcb5\"\n"
                + "            },\n"
                + "            {\n"
                + "              \"text\": \"Italian\",\n"
                + "              \"callback_data\": \"8768d660-f05f-4f4b-bda5-3451ab573d56\"\n"
                + "            }\n"
                + "          ]\n"
                + "        ]\n"
                + "      }\n"
                + "    }\n"
                + "  }\n"
                + "}",
            Update.class);
    final TgChat fakeTgChat = mock(TgChat.class);
    when(fakeTgChat.getId()).thenReturn(1111111L);
    when(fakeTgChat.getChatContext()).thenReturn(new ChatContext());
    when(fakeModelHelp.getMessage(eq(fakeTgChat), any())).thenReturn("doesn't matter");
    when(fakeModelHelp.generateInlineKeyBoardButton(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn(new InlineKeyboardButton[] {});
    when(fakeChatUtil.extractChat(eq(fakeCallbackQueryContext), eq(update)))
        .thenReturn(Optional.of(fakeTgChat));

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotResponseFuture =
        showHelp.process(fakeCallbackQueryContext, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = gotResponseFuture.get();
    final BaseRequest<?, ?> gotResponse = gotResponseOpt.get();
    final InlineKeyboardMarkup replyMarkup =
        (InlineKeyboardMarkup) gotResponse.getParameters().get("reply_markup");
    assertNull(replyMarkup);
  }
}
