package com.github.polpetta.mezzotre.telegram.command;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.util.Clock;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import io.ebean.Database;
import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import java.util.stream.Stream;
import org.apache.velocity.app.VelocityEngine;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Tag("velocity")
@Testcontainers
class HelpIntegrationTest {
  private static Gson gson;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private Database database;
  private VelocityEngine velocityEngine;
  private Clock fakeClock;
  private Help help;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));
    velocityEngine = Loader.defaultVelocityEngine();

    final Logger log = LoggerFactory.getLogger(Start.class);
    fakeClock = mock(Clock.class);
  }

  @Test
  void shouldProvideMessageWithButtons() throws Exception {
    when(fakeClock.now()).thenReturn(42L);

    final TgChat tgChat = new TgChat(1111111L, new ChatContext(), "en-US", false);
    tgChat.save();

    final HashMap<String, Processor> commands = new HashMap<>();
    final Processor dummy1 =
        new Processor() {
          @Override
          public Set<String> getTriggerKeywords() {
            return Set.of("/a", "/b");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              TgChat chat, Update update) {
            return null;
          }

          @Override
          public String getLocaleDescriptionKeyword() {
            return "start.cmdDescription";
          }
        };
    final Processor dummy2 =
        new Processor() {
          @Override
          public Set<String> getTriggerKeywords() {
            return Set.of("/different");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              TgChat chat, Update update) {
            return null;
          }

          @Override
          public String getLocaleDescriptionKeyword() {
            return "help.cmdDescription";
          }
        };
    commands.put("/a", dummy1);
    commands.put("/b", dummy1);
    commands.put("/different", dummy2);

    final Map<String, com.github.polpetta.mezzotre.telegram.callbackquery.Processor> events =
        new HashMap<>();

    final com.github.polpetta.mezzotre.telegram.callbackquery.Processor dummyEvent1 =
        new com.github.polpetta.mezzotre.telegram.callbackquery.Processor() {
          @Override
          public String getEventName() {
            return "exampleEvent";
          }

          @Override
          public boolean canBeDirectlyInvokedByTheUser() {
            return true;
          }

          @Override
          public Optional<String> getPrettyPrintLocaleKeyName() {
            return Optional.of("changeLanguage.inlineKeyboardButtonName");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              CallbackQueryContext callbackQueryContext, Update update) {
            return null;
          }
        };

    final com.github.polpetta.mezzotre.telegram.callbackquery.Processor dummyEvent2 =
        new com.github.polpetta.mezzotre.telegram.callbackquery.Processor() {
          @Override
          public String getEventName() {
            return "secondExampleEvent";
          }

          @Override
          public boolean canBeDirectlyInvokedByTheUser() {
            return true;
          }

          @Override
          public Optional<String> getPrettyPrintLocaleKeyName() {
            return Optional.of("selectLanguageTutorial.inlineKeyboardButtonName");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              CallbackQueryContext callbackQueryContext, Update update) {
            return null;
          }
        };

    events.put(dummyEvent1.getEventName(), dummyEvent1);
    events.put(dummyEvent2.getEventName(), dummyEvent2);

    final com.github.polpetta.mezzotre.telegram.model.Help modelHelp =
        new com.github.polpetta.mezzotre.telegram.model.Help(
            new TemplateContentGenerator(new LocalizedMessageFactory(velocityEngine)),
            new UUIDGenerator());

    help =
        new Help(
            Executors.newSingleThreadExecutor(),
            commands,
            events,
            modelHelp,
            new ChatUtil(fakeClock));

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/help\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotFuture = help.process(tgChat, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = gotFuture.get();
    final BaseRequest<?, ?> gotResponse = gotResponseOpt.get();
    assertInstanceOf(SendMessage.class, gotResponse);
    final String message = (String) gotResponse.getParameters().get("text");
    assertEquals(
        "Here is a list of what I can do:\n"
            + "\n"
            + "- /a /b: Trigger this very bot\n"
            + "- /different: Print the help message\n"
            + "\n"
            + "You can do the same operations you'd do with the commands aforementioned by"
            + " selecting the corresponding button below \uD83D\uDC47",
        message);

    final InlineKeyboardButton[][] keyboard =
        ((InlineKeyboardMarkup)
                gotResponse
                    .getParameters()
                    .getOrDefault("reply_markup", new InlineKeyboardMarkup()))
            .inlineKeyboard();

    final List<InlineKeyboardButton> keyboardButtons =
        Stream.of(keyboard).flatMap(Stream::of).toList();

    assertEquals(2, keyboardButtons.size());

    assertFalse(keyboardButtons.get(0).callbackData().isBlank());
    assertFalse(keyboardButtons.get(1).callbackData().isBlank());

    assertTrue(
        keyboardButtons.stream()
            .map(InlineKeyboardButton::text)
            .anyMatch("Select language"::equals));

    assertTrue(
        keyboardButtons.stream()
            .map(InlineKeyboardButton::text)
            .anyMatch("Change language"::equals));

    final TgChat gotChat = new QTgChat().id.eq(1111111L).findOne();
    assertNotNull(gotChat);
    assertTrue(gotChat.getHasHelpBeenShown());
    final ChatContext gotChatChatContext = gotChat.getChatContext();
    assertEquals(0, gotChatChatContext.getStep());
    assertEquals("/help", gotChatChatContext.getStage());
    assertEquals(42, gotChatChatContext.getPreviousMessageUnixTimestampInSeconds());
  }
}
