package com.github.polpetta.mezzotre.telegram.command;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.request.BaseRequest;
import java.util.Collections;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.parallel.Execution;
import org.junit.jupiter.api.parallel.ExecutionMode;

@Execution(ExecutionMode.CONCURRENT)
class HelpTest {

  private static Gson gson;
  private com.github.polpetta.mezzotre.telegram.model.Help fakeModelHelp;
  private ChatUtil fakeChatUtil;
  private Help help;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() {

    final Map<String, Processor> commands = Collections.emptyMap();
    final Map<String, com.github.polpetta.mezzotre.telegram.callbackquery.Processor> events =
        Collections.emptyMap();

    fakeModelHelp = mock(com.github.polpetta.mezzotre.telegram.model.Help.class);
    fakeChatUtil = mock(ChatUtil.class);

    help =
        new Help(
            Executors.newSingleThreadExecutor(), commands, events, fakeModelHelp, fakeChatUtil);
  }

  @Test
  void shouldUpdateChatContext() throws Exception {
    final TgChat fakeTgChat = mock(TgChat.class);
    when(fakeTgChat.getId()).thenReturn(1111111L);
    when(fakeTgChat.getLocale()).thenReturn("en-US");
    when(fakeTgChat.getChatContext()).thenReturn(new ChatContext());
    when(fakeModelHelp.getMessage(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn("doesn't matter");
    when(fakeModelHelp.generateInlineKeyBoardButton(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn(new InlineKeyboardButton[] {});

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/help\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotResponseFuture =
        help.process(fakeTgChat, update);
    assertDoesNotThrow(() -> gotResponseFuture.get());
    verify(fakeChatUtil, times(1))
        .updateChatContext(eq(fakeTgChat), eq("/help"), eq(0), eq(Collections.emptyMap()));
  }

  @Test
  void shouldAppendKeyboardIfButtonsAreReturned() throws Exception {
    final TgChat fakeTgChat = mock(TgChat.class);
    when(fakeTgChat.getId()).thenReturn(1111111L);
    when(fakeTgChat.getLocale()).thenReturn("en-US");
    when(fakeTgChat.getChatContext()).thenReturn(new ChatContext());
    when(fakeModelHelp.getMessage(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn("doesn't matter");
    when(fakeModelHelp.generateInlineKeyBoardButton(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn(new InlineKeyboardButton[] {new InlineKeyboardButton("example1")});

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/help\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotResponseFuture =
        help.process(fakeTgChat, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = gotResponseFuture.get();
    final BaseRequest<?, ?> gotResponse = gotResponseOpt.get();
    final InlineKeyboardMarkup replyMarkup =
        (InlineKeyboardMarkup) gotResponse.getParameters().get("reply_markup");
    assertEquals(1, replyMarkup.inlineKeyboard()[0].length);
    assertEquals("example1", replyMarkup.inlineKeyboard()[0][0].text());
  }

  @Test
  void shouldNotAppendKeyboardIfNoButtonIsPresent() throws Exception {
    final TgChat fakeTgChat = mock(TgChat.class);
    when(fakeTgChat.getId()).thenReturn(1111111L);
    when(fakeTgChat.getLocale()).thenReturn("en-US");
    when(fakeTgChat.getChatContext()).thenReturn(new ChatContext());
    when(fakeModelHelp.getMessage(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn("doesn't matter");
    when(fakeModelHelp.generateInlineKeyBoardButton(eq(fakeTgChat), eq(Collections.emptyMap())))
        .thenReturn(new InlineKeyboardButton[] {});

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/help\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotResponseFuture =
        help.process(fakeTgChat, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = gotResponseFuture.get();
    final BaseRequest<?, ?> gotResponse = gotResponseOpt.get();
    final InlineKeyboardMarkup replyMarkup =
        (InlineKeyboardMarkup) gotResponse.getParameters().get("reply_markup");
    assertNull(replyMarkup);
  }
}
