package com.github.polpetta.mezzotre.telegram.command;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.util.Clock;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import io.ebean.Database;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import org.apache.velocity.app.VelocityEngine;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Tag("velocity")
@Testcontainers
class StartIntegrationTest {

  private static Gson gson;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private VelocityEngine velocityEngine;
  private LocalizedMessageFactory localizedMessageFactory;
  private Start start;
  private Database database;
  private Clock fakeClock;
  private UUIDGenerator fakeUUIDGenerator;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));
    velocityEngine = Loader.defaultVelocityEngine();

    final Logger log = LoggerFactory.getLogger(Start.class);

    fakeClock = mock(Clock.class);
    fakeUUIDGenerator = mock(UUIDGenerator.class);

    start =
        new Start(
            new TemplateContentGenerator(new LocalizedMessageFactory(velocityEngine)),
            Executors.newSingleThreadExecutor(),
            log,
            fakeUUIDGenerator,
            "Mezzotre",
            new ChatUtil(fakeClock));
  }

  @Test
  void shouldUpdateContextInTheDatabase() throws Exception {
    when(fakeClock.now()).thenReturn(42L);
    when(fakeUUIDGenerator.generateAsString())
        .thenReturn("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5")
        .thenReturn("16507fbd-9f28-48a8-9de1-3ea1c943af67")
        .thenReturn("0b0ac18e-f621-484e-aa8d-9b176be5b930");
    final TgChat tgChat = new TgChat(1111111L, new ChatContext());
    tgChat.setLocale("en-US");
    tgChat.save();

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/start\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotFuture = start.process(tgChat, update);
    assertDoesNotThrow(() -> gotFuture.get());
    final Optional<BaseRequest<?, ?>> gotMessageOptional = gotFuture.get();
    assertDoesNotThrow(gotMessageOptional::get);
    final BaseRequest<?, ?> gotMessage = gotMessageOptional.get();
    assertInstanceOf(SendMessage.class, gotMessage);
    final String message = (String) gotMessage.getParameters().get("text");
    assertEquals(
        "*Hello Test Firstname! \uD83D\uDC4B*\n\n"
            + "This is _Mezzotre_, a simple bot focused on DnD content management! Please start by"
            + " choosing a language down below \uD83D\uDC47",
        message);
    assertEquals(1111111L, (Long) gotMessage.getParameters().get("chat_id"));

    final TgChat retrievedTgChat = new QTgChat().id.eq(1111111L).findOne();
    assertNotNull(retrievedTgChat);
    final ChatContext gotChatContext = retrievedTgChat.getChatContext();
    assertEquals(42, gotChatContext.getPreviousMessageUnixTimestampInSeconds());
    assertEquals(0, gotChatContext.getLastMessageSentId());
    assertEquals("/start", gotChatContext.getStage());
    assertEquals(0, gotChatContext.getStep());
  }

  @Test
  void shouldReceiveHelloIntroduction() throws Exception {
    when(fakeClock.now()).thenReturn(42L);
    when(fakeUUIDGenerator.generateAsString())
        .thenReturn("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5")
        .thenReturn("16507fbd-9f28-48a8-9de1-3ea1c943af67")
        .thenReturn("0b0ac18e-f621-484e-aa8d-9b176be5b930");

    final TgChat tgChat = new TgChat(1111111L, new ChatContext(), "en-US", false);

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/start\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotFuture = start.process(tgChat, update);
    assertDoesNotThrow(() -> gotFuture.get());
    final Optional<BaseRequest<?, ?>> gotMessageOptional = gotFuture.get();
    assertDoesNotThrow(gotMessageOptional::get);
    final BaseRequest<?, ?> gotMessage = gotMessageOptional.get();
    assertInstanceOf(SendMessage.class, gotMessage);
    final String message = (String) gotMessage.getParameters().get("text");
    assertEquals(
        "*Hello Test Firstname! \uD83D\uDC4B*\n\n"
            + "This is _Mezzotre_, a simple bot focused on DnD content management! Please start by"
            + " choosing a language down below \uD83D\uDC47",
        message);
    assertEquals(1111111L, (Long) gotMessage.getParameters().get("chat_id"));

    verify(fakeUUIDGenerator, times(3)).generateAsString();
    assertEquals(
        2,
        new QCallbackQueryContext()
            .entryGroup
            .eq("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5")
            .findCount());
    assertEquals(2, new QCallbackQueryContext().findCount());
  }

  @Test
  void shouldThrowErrorIfLocaleNonExists() {
    when(fakeClock.now()).thenReturn(42L);
    when(fakeUUIDGenerator.generateAsString())
        .thenReturn("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5")
        .thenReturn("16507fbd-9f28-48a8-9de1-3ea1c943af67")
        .thenReturn("0b0ac18e-f621-484e-aa8d-9b176be5b930");
    final TgChat tgChat = new TgChat(1111111L, null);

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"message\":{\n"
                + "  \"date\":1441645532,\n"
                + "  \"chat\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"type\": \"private\",\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"message_id\":1365,\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"text\":\"/start\"\n"
                + "}\n"
                + "}",
            Update.class);

    final CompletableFuture<Optional<BaseRequest<?, ?>>> gotFuture = start.process(tgChat, update);
    assertThrows(ExecutionException.class, gotFuture::get);
  }
}
