package com.github.polpetta.mezzotre.telegram.model;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.mezzotre.telegram.callbackquery.Processor;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.ChatContext;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.request.BaseRequest;
import io.ebean.Database;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Tag("velocity")
@Testcontainers
class HelpIntegrationTest {

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private TemplateContentGenerator templateContentGenerator;
  private UUIDGenerator fakeUUIDGenerator;
  private Help help;
  private Database database;

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));
    templateContentGenerator =
        new TemplateContentGenerator(new LocalizedMessageFactory(Loader.defaultVelocityEngine()));
    fakeUUIDGenerator = mock(UUIDGenerator.class);

    help = new Help(templateContentGenerator, fakeUUIDGenerator);
  }

  @Test
  void shouldGenerateAGoodHelpMessage() {
    final TgChat tgChat = new TgChat(11111L, new ChatContext());
    tgChat.save();

    final com.github.polpetta.mezzotre.telegram.command.Processor dummyCommand1 =
        new com.github.polpetta.mezzotre.telegram.command.Processor() {
          @Override
          public Set<String> getTriggerKeywords() {
            return Set.of("/example", "/another");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              TgChat chat, Update update) {
            return CompletableFuture.completedFuture(Optional.empty());
          }

          @Override
          public String getLocaleDescriptionKeyword() {
            return "help.cmdDescription";
          }
        };

    final String gotMessage =
        help.getMessage(tgChat, Map.of("/example", dummyCommand1, "/another", dummyCommand1));
    assertEquals(
        "Here is a list of what I can do:\n"
            + "\n"
            + "- /another /example: Print the help message\n"
            + "\n"
            + "You can do the same operations you'd do with the commands aforementioned by"
            + " selecting the corresponding button below \uD83D\uDC47",
        gotMessage);
  }

  @Test
  void shouldGenerateButtonsCorrectly() {
    final TgChat tgChat = new TgChat(111111L, new ChatContext());
    tgChat.save();
    when(fakeUUIDGenerator.generateAsString())
        .thenReturn("53dc6dca-1042-4bc7-beb8-ce2a34df6a54")
        .thenReturn("d5d3e016-7b60-4f1a-bd79-e1a6bff32f17");

    final Processor visibleProcessor1 =
        new Processor() {
          @Override
          public String getEventName() {
            return "eventName";
          }

          @Override
          public boolean canBeDirectlyInvokedByTheUser() {
            return true;
          }

          @Override
          public Optional<String> getPrettyPrintLocaleKeyName() {
            return Optional.of("changeLanguage.inlineKeyboardButtonName");
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              CallbackQueryContext callbackQueryContext, Update update) {
            return CompletableFuture.completedFuture(Optional.empty());
          }
        };

    final Processor invisibleProcessor1 =
        new Processor() {
          @Override
          public String getEventName() {
            return "invisible";
          }

          @Override
          public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
              CallbackQueryContext callbackQueryContext, Update update) {
            return CompletableFuture.completedFuture(Optional.empty());
          }
        };

    final InlineKeyboardButton[] buttons =
        help.generateInlineKeyBoardButton(
            tgChat,
            Map.of(
                visibleProcessor1.getEventName(),
                visibleProcessor1,
                invisibleProcessor1.getEventName(),
                invisibleProcessor1));

    assertEquals(1, buttons.length);
    assertEquals("Change language", buttons[0].text());

    assertEquals(1, new QCallbackQueryContext().findCount());
  }
}
